# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from ._jensen_shannon_divergence import jensen_shannon_divergence

__all__ = ["jensen_shannon_divergence"]
